/** \file
 *  \brief Main form module
 */

/* Copyright (C) 2008-2009 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 */

#include <vcl.h>
#pragma hdrstop

#include "Main.h"
#include "display\DisplayOpenGL.h"
#include "display\ValueStringFormatter.h"
#include "FFT.h"
#include "Log.h"
#include "LogUnit.h"
#include "NotificationUnit.h"
#include "ScopedLock.h"
#include "Mutex.h"
#include "AboutUnit.h"
#include "MyTrackBar.h"

#include "device\MiniscopeDeviceInterface.h"

#include "SettingsUnit.h"
#include "Settings.h"

#include "Recorder.h"

#include "StatisticsUnit.h"
#include "CalibrateSensitivityUnit.h"
#include "PlotToolbarUnit.h"
#include "WaitUnit.h"
#include "JsonViewerUnit.h"
#include "Ms3ViewerUnit.h"
#include "Ms3Recorder.h"
#include "Ms3FileRepairUnit.h"
#include "FilterSelector.h"
#include "DataFilterChain.h"

#include "TimeCounter.h"
#include "CompileTimeChecker.h"

#include <SysUtils.hpp>
#include <ClipBrd.hpp>
#include <math.h>
#include <limits>
#include <algorithm>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TfrmMain *frmMain;
Settings appSettings;   ///< application settings instance
DataFilterChain filterChain;

//---------------------------------------------------------------------------
__fastcall TfrmMain::TfrmMain(TComponent* Owner)
	: TForm(Owner), fZoomChangeGrid(1.25f), device(NULL), bPaintUpdate(false),
	bFrameFilled(false),
	bFrameProgressUpdate(false),
	bJsonViewerMode(false),
	bMs3ViewerMode(false)
{
	CLog::Instance()->SetFile(ChangeFileExt(Application->ExeName, ".log").c_str());
	CLog::Instance()->SetLevel(E_LOG_TRACE);

	display = new DisplayOpenGL(pnlCenter->Handle);
	// some random settings
	display->SetDiv(0.01f);	// FS = 100Hz
	display->SetRange(1.0);	
    display->data = &dataOsc;
	btnModeScope->Down = true;

	display->ChangeSize(pnlCenter->Width, pnlCenter->Height);

	display->SetMode(Display::MODE_OSCILLOSCOPE);		
	iStoredTriggerTypePos = -1;	

	// load custom cursors
	Screen->Cursors[CURSOR_HAND] = LoadCursor(HInstance, "HANDCUR");
	Screen->Cursors[CURSOR_ZOOM] = LoadCursor(HInstance, "ZOOMCUR");

	mouse_tool = MOUSE_CURSOR;
	pnlCenter->Cursor = crCross;
	btnMouseCursor->Down = true;

	bMovingPlot = false;

	displayOsc.fZoomX = 1.0;
	displayOsc.fZoomY = 1.0;
	displaySpe.fZoomX = 1.0;
	displaySpe.fZoomY = 2.0;
	displayRec.fZoomX = 1.0;
	displayRec.fZoomY = 1.0;
	UpdateZoomDisplay();

	float fValX, fValY;
	display->SetCursor1(0, pnlCenter->Height/2);
	display->PosToPhysical(0, pnlCenter->Height/2, fValX, fValY);
	UpdateCursorReadings(Display::MODE_OSCILLOSCOPE, 1, fValX, fValY);
	display->SetCursor2(0, pnlCenter->Height/2);
	display->PosToPhysical(0, pnlCenter->Height/2, fValX, fValY);
	UpdateCursorReadings(Display::MODE_OSCILLOSCOPE, 2, fValX, fValY);

	bTriggered = false;
	memset(&current, 0 , sizeof(current));

	Timer->Interval = 50;

	//STATIC_CHECK(sizeof(int) <= sizeof(char), Just_Testing);
}
//---------------------------------------------------------------------------

__fastcall TfrmMain::~TfrmMain()
{
	FFT::SetWindowType("");	// cleanup (makes CG happy)
	CLog::Instance()->Destroy();
}

void __fastcall TfrmMain::pnlCenterResize(TObject *Sender)
{
	if (display)
		display->ChangeSize(pnlCenter->Width, pnlCenter->Height);
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::TimerTimer(TObject *Sender)
{
	Timer->Enabled = false;

	ScopedLock<Mutex> lock(mutexDeviceData);

	if (bTriggered)
	{
		ProgressBarFrameRx->Visible = true;
	}
    unsigned int iExpectedFrameSize = current.buffer_size;
	if (bDataUpdate)
	{
		if (dataOsc.size() >= iExpectedFrameSize)
			bFrameFilled = true;
		else
			bFrameFilled = false;
		filterChain.Process(dataOsc);
		if (display->GetMode() == Display::MODE_SPECTRUM)
		{
			if (bFrameFilled)
			{
				if (frmStatistics->Visible)
					frmStatistics->Process(dataOsc, current.fDivX);
				CalculateSpectrum();
			}
		}
		else if (display->GetMode() == Display::MODE_RECORDER)
		{
			if (bFrameFilled)
			{
				if (frmStatistics->Visible)
					frmStatistics->Process(dataRec, current.fDivX);
				std::vector<float>::const_iterator it;
				float fAvg = 0;
				if (dataOsc.size() > 0)
				{
					for (it = dataOsc.begin(); it != dataOsc.end(); ++it)
					{
						fAvg += (*it);
					}
					fAvg /= dataOsc.size();
				}
				dataRec.push_back(fAvg);
			}
		}
		else if (display->GetMode() == Display::MODE_OSCILLOSCOPE)
		{
			if (frmStatistics->Visible)
				frmStatistics->Process(dataOsc, current.fDivX);
		}
	}

	if (bFrameProgressUpdate || bDataUpdate)
	{
		if (iExpectedFrameSize == 0)
		{
			ProgressBarFrameRx->Position == ProgressBarFrameRx->Min;
		}
		else
		{
			if (bFrameFilled)
				ProgressBarFrameRx->Position = 100*dataOscTmp.size()/iExpectedFrameSize;
			else
				ProgressBarFrameRx->Position = 100*dataOsc.size()/iExpectedFrameSize;
		}

		if (ProgressBarFrameRx->Position == ProgressBarFrameRx->Max)
		{
			ProgressBarFrameRx->Position = ProgressBarFrameRx->Min;
			ProgressBarFrameRx->Visible = false;
			bTriggered = false;
		}
		//if (ProgressBarFrameRx->Position < 20 || ProgressBarFrameRx->Position == 100)
		//	LOG(E_LOG_TRACE, "Progress: %d", ProgressBarFrameRx->Position);
		ProgressBarFrameRx->Update();
		bFrameProgressUpdate = false;
	}

	if (bDataUpdate || bPaintUpdate)
	{
		//TimeCounter tc("Display update");
		display->Update(Display::UPDATE_ALL);
		bPaintUpdate = false;
		bDataUpdate = false;
	}
	Timer->Enabled = true;
}
//---------------------------------------------------------------------------

void TfrmMain::SetCapabilities(const struct S_SCOPE_CAPABILITIES* const capabilities)
{
	//capabilities->iChannelCount
	if (capabilities->iSensitivityLevelsCount > 1) {
		lblSensitivity->Visible = true;
		lblSensitivityVal->Visible = true;
		trbarSensitivity->Visible = true;
	} else {
		lblSensitivity->Visible = false;
		lblSensitivityVal->Visible = false;
		trbarSensitivity->Visible = false;
	}
	trbarSensitivity->Max = capabilities->iSensitivityLevelsCount - 1;
	trbarSensitivity->Position = trbarSensitivity->Max;

	if (capabilities->iSamplingPeriodCount > 1) {
		lblSampling->Visible = true;
		lblSamplingVal->Visible = true;
		trbarSampling->Visible = true;
	} else {
		lblSampling->Visible = false;
		lblSamplingVal->Visible = false;
		trbarSampling->Visible = false;
	}
	trbarSampling->Max = capabilities->iSamplingPeriodCount - 1;
	trbarSampling->Position = trbarSampling->Max;

	//trbarTriggerLevel->Min = pow(2, capabilities->iBitsPerSample)/(-2);
	//trbarTriggerLevel->Max = pow(2, capabilities->iBitsPerSample)/2 - 1;
	trbarTriggerLevel->Min = -100;
	trbarTriggerLevel->Max = 100;
	trbarTriggerLevel->Position = 0;

	this->cbTriggerType->Clear();
	ActionModeRecorder->Enabled = false;
	for (unsigned int i=0; i<capabilities->iTriggerTypeCount; i++) {
		AnsiString name = DeviceInterface::GetTriggerTypeName(capabilities->iTriggerTypes[i]);
		switch (capabilities->iTriggerTypes[i]) {
		case TRIGGER_MANUAL:
			cbTriggerType->Items->Add(name);
			ActionModeRecorder->Enabled = true;
			break;
		case TRIGGER_SLOPE_UP:
			cbTriggerType->Items->Add(name);
			break;
		case TRIGGER_SLOPE_DOWN:
			cbTriggerType->Items->Add(name);
			break;
		case TRIGGER_CONTINUOUS:
			cbTriggerType->Items->Add(name);
			break;
		default:
			break;
		}
	}
	if (capabilities->iTriggerTypeCount > 0) {
		cbTriggerType->ItemIndex = 0;
	}

	this->cbCouplingType->Clear();
	for (unsigned int i=0; i<capabilities->iCouplingTypeCount; i++) {
		AnsiString name = DeviceInterface::GetCouplingTypeName(capabilities->iCouplingType[i]);
		switch (capabilities->iCouplingType[i]) {
		case COUPLING_GND:
			cbCouplingType->Items->Add(name);
			break;
		case COUPLING_DC:
			cbCouplingType->Items->Add(name);
			break;
		case COUPLING_AC:
			cbCouplingType->Items->Add(name);
			break;
		default:
			break;
		}
	}
	if (capabilities->iCouplingTypeCount > 0) {
		cbCouplingType->ItemIndex = 0;
	}

	this->cbBufferSize->Clear();
	for (unsigned int i=0; i<capabilities->iBufferSizeCount; i++) {
		cbBufferSize->Items->Add(capabilities->iBufferSizes[i]);
	}
	if (capabilities->iBufferSizeCount > 0) {
    	cbBufferSize->ItemIndex = 0;
	}
}
//---------------------------------------------------------------------------

void TfrmMain::CalculateSpectrum(void)
{
	// FFT
	unsigned int count = FFT::NearestPowerOfTwo(dataOsc.size());
	if (count > appSettings.FFT.iSamplesLimit)
		count = appSettings.FFT.iSamplesLimit;
	FFT::SetSamplesCount(count);
	float *ReIn = new float[count];
	float *ReOut = new float[count];
	float *ImOut = new float[count];
	for(unsigned int i=0;i < count;i++)
		ReIn[i]= dataOsc[i];
	FFT::Transform(ReIn, ReOut, ImOut);
	dataFFT.resize(count/2);
	//LOG(E_LOG_TRACE, "started");

	for(unsigned int i=0;i < count/2;i++)
	{
		dataFFT[i] = sqrt((ReOut[i]*ReOut[i])+(ImOut[i]*ImOut[i]))/count * 2;
		//LOG(E_LOG_TRACE, "%f", dataFFT[i]);
	}
	if (count > 0) {
    	dataFFT[0] /= 2.0f;
	}
	delete [] ReIn;
	delete [] ReOut;
	delete [] ImOut;
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::pnlCenterMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
	if (Button == mbMiddle)
	{
		if (!frmPlotToolbar->Visible) {
			POINT P = ClientToScreen(Point(pnlCenter->Left + X, pnlCenter->Top + Y));
			frmPlotToolbar->ShowAt(P.x, P.y);
		} else
			frmPlotToolbar->Hide();
		return;
	}

	if (mouse_tool == MOUSE_CURSOR )
	{
		float fValX, fValY;
		if(Button==mbLeft)  //left button clicked
		{
			display->SetCursor1(X, Y);
			display->PosToPhysical(X, Y, fValX, fValY);
			UpdateCursorReadings(display->GetMode(), 1, fValX, fValY);
		}
		else if(Button==mbRight)
		{
			display->SetCursor2(X, Y);
			display->PosToPhysical(X, Y, fValX, fValY);
			UpdateCursorReadings(display->GetMode(), 2, fValX, fValY);
		}
	}
	else if (mouse_tool == MOUSE_OFFSET)
	{
		iPlotMoveStartX = X;
		iPlotMoveStartY = Y;
		bMovingPlot = true;
	}
	else if (mouse_tool == MOUSE_ZOOM || mouse_tool == MOUSE_ZOOM_X || mouse_tool == MOUSE_ZOOM_Y)
	{
		if(Button==mbLeft)  //left button clicked
		{
			if (mouse_tool != MOUSE_ZOOM_Y)
				DisplayState().fZoomX *= fZoomChangeGrid;
			if (mouse_tool != MOUSE_ZOOM_X)
				DisplayState().fZoomY *= fZoomChangeGrid;
		}
		else if(Button==mbRight)
		{
			if (mouse_tool != MOUSE_ZOOM_Y)
				DisplayState().fZoomX /= fZoomChangeGrid;
			if (mouse_tool != MOUSE_ZOOM_X)
				DisplayState().fZoomY /= fZoomChangeGrid;
		}
		display->SetZoom((mouse_tool==MOUSE_ZOOM_Y)?(Display::OFFSET_LEAVE):(X),
			(mouse_tool==MOUSE_ZOOM_X)?(Display::OFFSET_LEAVE):(Y),
			DisplayState().fZoomX, DisplayState().fZoomY);
		UpdateZoomDisplay();
	}
	bPaintUpdate = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::pnlCenterMouseUp(TObject *Sender, TMouseButton Button,
	  TShiftState Shift, int X, int Y)
{
	if (bMovingPlot)
	{
		display->Move(iPlotMoveStartX, iPlotMoveStartY, X, Y);
		bMovingPlot = false;
		bPaintUpdate = true;
	}
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateCursorReadings(enum Display::E_DISPLAY_MODE mode, int cursorId, float fValX, float fValY)
{
	static float fValX1 = 0.0, fValY1 = 0.0, fValX2 = 0.0, fValY2 = 0.0;

	if (mode==Display::MODE_OSCILLOSCOPE || mode==Display::MODE_RECORDER)
	{
		if (cursorId==1)
		{
			StatusBar->Panels->Items[0]->Text =
				(ValueStringFormatter::TimeToStr(fValX, 3) + " " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX1 = fValX;
			fValY1 = fValY;
		}
		else if (cursorId==2)
		{
			StatusBar->Panels->Items[1]->Text =
				(ValueStringFormatter::TimeToStr(fValX, 3) + " " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX2 = fValX;
			fValY2 = fValY;
		}
		float dx = fabs(fValX1-fValX2);
		float dy = fabs(fValY1-fValY2);
		StatusBar->Panels->Items[2]->Text =
			((std::string)"dx=" + ValueStringFormatter::TimeToStr(dx, 3) + " dy=" +
			ValueStringFormatter::VoltageToStr(dy, 3)).c_str();
		if (dx != 0)
		{
			StatusBar->Panels->Items[2]->Text = StatusBar->Panels->Items[2]->Text +
				" 1/dx=" + ValueStringFormatter::FreqToStr(1/dx, 3).c_str();
		}
	}
	else
	{
		if (cursorId==1)
		{
			StatusBar->Panels->Items[0]->Text =
				(ValueStringFormatter::FreqToStr(fValX, 3) + "  " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX1 = fValX;
			fValY1 = fValY;
		}
		else if (cursorId==2)
		{
			StatusBar->Panels->Items[1]->Text =
				(ValueStringFormatter::FreqToStr(fValX, 3) + "  " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX2 = fValX;
			fValY2 = fValY;
		}
		float dx = fabs(fValX1-fValX2);
		float dy = fabs(fValY1-fValY2);
		StatusBar->Panels->Items[2]->Text =
			((std::string)"dx=" + ValueStringFormatter::FreqToStr(dx, 3) + "  dy=" +
			ValueStringFormatter::VoltageToStr(dy, 3)).c_str();
	}
}


void __fastcall TfrmMain::ActionChangeMouseModeCursorExecute(TObject *Sender)
{
	mouse_tool = MOUSE_CURSOR;
	pnlCenter->Cursor = crCross;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeOffsetExecute(TObject *Sender)
{
	mouse_tool = MOUSE_OFFSET;
	pnlCenter->Cursor = (TCursor)CURSOR_HAND;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeZoomExecute(TObject *Sender)
{
	mouse_tool = MOUSE_ZOOM;
	pnlCenter->Cursor = (TCursor)CURSOR_ZOOM;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeZoomXExecute(TObject *Sender)
{
	mouse_tool = MOUSE_ZOOM_X;
	pnlCenter->Cursor = (TCursor)CURSOR_ZOOM;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeZoomYExecute(TObject *Sender)
{
	mouse_tool = MOUSE_ZOOM_Y;
	pnlCenter->Cursor = (TCursor)CURSOR_ZOOM;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionZoomClearExecute(TObject *Sender)
{
	DisplayState().fZoomX = 1.0;
	if (display->GetMode() == Display::MODE_SPECTRUM)
		DisplayState().fZoomY = 2.0;
	else
		DisplayState().fZoomY = 1.0;
	display->SetZoom(Display::OFFSET_CLEAR, Display::OFFSET_CLEAR,
		DisplayState().fZoomX, DisplayState().fZoomY);
	UpdateZoomDisplay();
	bPaintUpdate = true;	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionModeScopeExecute(TObject *Sender)
{
	SetMode(Display::MODE_OSCILLOSCOPE);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionModeSpectrumExecute(TObject *Sender)
{
	SetMode(Display::MODE_SPECTRUM);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionModeRecorderExecute(TObject *Sender)
{
	SetMode(Display::MODE_RECORDER);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miAboutClick(TObject *Sender)
{
	AboutBox->ShowModal();
}
//---------------------------------------------------------------------------



void __fastcall TfrmMain::FormCreate(TObject *Sender)
{
    AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".ini" );
	appSettings.Read(asConfigFile);
	this->Width = appSettings.iWidth;
	this->Height = appSettings.iHeight;
	this->Top = appSettings.iPosY;
	this->Left = appSettings.iPosX;
	if (appSettings.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
	if (appSettings.bWindowMaximized)
		this->WindowState = wsMaximized;
	edInterval->Text = AnsiString::FloatToStrF(appSettings.fRecorderInterval, AnsiString::sffGeneral, 1, 3);
	FFT::SetWindowType(appSettings.FFT.asWindowName);
	((TMyTrackBar*)trbarSensitivity)->OnMouseUp = TrackBarMouseUp;
	((TMyTrackBar*)trbarSampling)->OnMouseUp = TrackBarMouseUp;
	((TMyTrackBar*)trbarTriggerLevel)->OnMouseUp = TrackBarMouseUp;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miDeviceSettingsClick(TObject *Sender)
{
		device->ShowSettings();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miViewLogClick(TObject *Sender)
{
	frmLog->Visible = miViewLog->Checked;
	if (frmLog->Visible)
		frmLog->BringToFront();
}
//---------------------------------------------------------------------------

void TfrmMain::CallbackCloseLogFn(void)
{
	miViewLog->Checked = false;
}

void TfrmMain::CallbackDataFn(short* buf, int size)
{
	/** \note DO NOT use VCL code inside callbacks called by device dll */
	ScopedLock<Mutex> lock(mutexDeviceData);
	size /= sizeof(*buf);
	//LOG(E_LOG_TRACE, "DataFn: %d samples, total: %d samples", size, dataOscTmp.size());
	if (bJsonViewerMode || bMs3ViewerMode)
		return;

	for (int i=0; i<size; i++)
	{
		dataOscTmp.push_back((buf[i]-current.calibration.offset) * current.fDivY * current.calibration.gain);
	}

	if (bFrameFilled)
	{
		if ((int)dataOscTmp.size() >= current.buffer_size)
		{
			dataOsc = dataOscTmp;
			dataOscTmp.clear();
			dataOscTmp.reserve(102400);
			bDataUpdate = true;
			if (Ms3Recorder::Instance().GetMode() == Ms3Recorder::MODE_WRITE)
			{
				Ms3Recorder::Instance().StoreFrame(current, dataOsc);
				if (Ms3Recorder::Instance().GetFileSize() > (unsigned int)appSettings.Ms3Recording.iSizeLimit*1024)
				{
					TfrmNotification::Add(TfrmNotification::WARNING, "MS3 file size reached (check settings), recording stopped");
					ActionMs3Record->Execute();
				}
			}
		}
		else
		{
			bFrameProgressUpdate = true;
		}
	}
	else
	{
		int end = dataOscTmp.size();
		for (int i=dataOsc.size(); i<end; i++)
		{
			dataOsc.push_back(dataOscTmp[i]);
		}
		if ((int)dataOscTmp.size() >= current.buffer_size)
		{
			dataOscTmp.clear();
			dataOscTmp.reserve(102400);
			bDataUpdate = true;
			if (Ms3Recorder::Instance().GetMode() == Ms3Recorder::MODE_WRITE)
				Ms3Recorder::Instance().StoreFrame(current, dataOsc);
		}
		bDataUpdate = true;
	}
}

void TfrmMain::CallbackTriggerFn(void)
{
	ScopedLock<Mutex> lock(mutexDeviceData);
	if (bJsonViewerMode || bMs3ViewerMode)
		return;
	/** \note DO NOT use VCL code inside callbacks called by device dll */

	if (current.fDivX != upcoming.fDivX ||
		current.fDivY != upcoming.fDivY ||
		current.coupling != upcoming.coupling ||
		current.trigger_type != upcoming.trigger_type ||
		current.buffer_size != upcoming.buffer_size)
	{
		// clear display if any acquisition settings changed
		dataOsc.clear();
		dataOsc.reserve(102400);
		bDataUpdate = true;
	}
	dataOscTmp.clear();
	dataOscTmp.reserve(102400);

	current = upcoming;
	ftime(&current.timestamp);
	bTriggered = true;
}

void TfrmMain::CallbackConnectFn(int state, char *szMsgText)
{
	tmrUpdateConnState->Enabled = true;
}

void TfrmMain::UpdateConnectionState(void)
{
	struct ConnectionInfo conn = device->GetConnectionInfo();
	pnlOscControl->Enabled = false;
	switch (conn.state)
	{
	case DEVICE_DISCONNECTED:
		break;
	case DEVICE_CONNECTING:
		break;
	case DEVICE_CONNECTED:
		pnlOscControl->Enabled = true;
		break;
	case DEVICE_DISCONNECTING:
		break;
	default:
		break;
	}
    StatusBar->Panels->Items[3]->Text = conn.msg;
}

void __fastcall TfrmMain::ActionAcquireStartExecute(TObject *Sender)
{
	device->AcquireStart();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionAcquireStopExecute(TObject *Sender)
{
	device->AcquireStop();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::tmrUpdateConnStateTimer(TObject *Sender)
{
	tmrUpdateConnState->Enabled = false;
	UpdateConnectionState();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::cbTriggerTypeChange(TObject *Sender)
{
	UpdateTriggerType(true);
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateTriggerType(bool bUpdateDevice)
{
	enum E_TRIGGER_TYPE type = device->capabilities->iTriggerTypes[cbTriggerType->ItemIndex];
	if (bUpdateDevice)
		device->SetTriggerType(cbTriggerType->ItemIndex);
	btnManualTrigger->Visible = false;
	trbarTriggerLevel->Visible = false;
	lblTriggerLevel->Visible = false;
	lblTriggerLevelVal->Visible = false;
	btnStart->Visible = false;
	btnStop->Visible = false;
	switch (type) {
	case TRIGGER_MANUAL:
		btnManualTrigger->Visible = true;
		break;
	case TRIGGER_SLOPE_UP:
		trbarTriggerLevel->Visible = true;
		lblTriggerLevel->Visible = true;
		lblTriggerLevelVal->Visible = true;
		break;
	case TRIGGER_SLOPE_DOWN:
		trbarTriggerLevel->Visible = true;
		lblTriggerLevel->Visible = true;
		lblTriggerLevelVal->Visible = true;
		break;
	case TRIGGER_CONTINUOUS:
		btnStart->Visible = true;
		btnStop->Visible = true;
		break;
	default:
		break;
	}
	upcoming.trigger_type = device->capabilities->iTriggerTypes[cbTriggerType->ItemIndex];
}

void TfrmMain::UpdateCouplingType(void)
{
	device->SetCouplingType(cbCouplingType->ItemIndex);
	upcoming.coupling = device->capabilities->iCouplingType[cbCouplingType->ItemIndex];
}

void __fastcall TfrmMain::btnManualTriggerClick(TObject *Sender)
{
	int status = device->ManualTrigger();
	if (status)
		LOG(E_LOG_TRACE, "ManualTrigger() failed with status %d", status);
	else
		ProgressBarFrameRx->Visible = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::cbBufferSizeChange(TObject *Sender)
{
	UpdateBufferSize();
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateBufferSize(void)
{
	if (!cbBufferSize->Items->Count)
		return;
	device->SetBufferSize(cbBufferSize->ItemIndex);
	upcoming.buffer_size = device->capabilities->iBufferSizes[cbBufferSize->ItemIndex];
}

void TfrmMain::UpdateSensitivity(bool bUpdateDevice)
{
	if (bUpdateDevice)
		device->SetSensitivity(trbarSensitivity->Position);
	// update range caption, lblSensitivityVal
	float fDivVal = (float)pow(2, device->capabilities->iBitsPerSample) / (float)10 *
		device->capabilities->fSensitivityLevels[trbarSensitivity->Position] *
		calibration[trbarSensitivity->Position].gain;
	lblSensitivityVal->Caption =
		ValueStringFormatter::VoltageToStr(fDivVal, 1).c_str();
	lblSensitivityVal->Caption = lblSensitivityVal->Caption + "/div";
	upcoming.fDivY = device->capabilities->fSensitivityLevels[trbarSensitivity->Position];
	upcoming.calibration = calibration[trbarSensitivity->Position];

	if (display)
	{
		display->SetRange(fDivVal * 5.0);
		display->SetDiv(device->capabilities->fSamplingPeriod[trbarSampling->Position]);
		bDataUpdate = true;
	}
	frmCalibrateSensitivity->SetCurrentCalibration(
		calibration[trbarSensitivity->Position].offset,
		calibration[trbarSensitivity->Position].gain);
}

void TfrmMain::UpdateSampling(bool bUpdateDevice)
{
	if (bUpdateDevice)
		device->SetSampling(trbarSampling->Position);
	float fSamplVal = (float)1.0 / device->capabilities->fSamplingPeriod[trbarSampling->Position];
	lblSamplingVal->Caption =
		ValueStringFormatter::FreqToStr(fSamplVal, 1).c_str();
	upcoming.fDivX = device->capabilities->fSamplingPeriod[trbarSampling->Position];
	if (display)
	{
		display->SetDiv(device->capabilities->fSamplingPeriod[trbarSampling->Position]);
	}
}

void TfrmMain::UpdateTriggerLevel(void)
{
	device->SetTriggerLevel(trbarTriggerLevel->Position);
	lblTriggerLevelVal->Caption = IntToStr(trbarTriggerLevel->Position) + "%";
}

void __fastcall TfrmMain::cbCouplingTypeChange(TObject *Sender)
{
	UpdateCouplingType();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::PaintBoxCenterPaint(TObject *Sender)
{
	/** \note PaintBoxCenter purpose is to fire OnPaint event
		to request graph repaint (TPanel has no OnPaint event but
		TPaintBox has no Handle)
    */
	bPaintUpdate = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::PaintBoxCenterMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
	pnlCenterMouseDown(Sender, Button, Shift, X, Y);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::PaintBoxCenterMouseUp(TObject *Sender,
	  TMouseButton Button, TShiftState Shift, int X, int Y)
{
	pnlCenterMouseUp(Sender, Button, Shift, X, Y);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::FormCloseQuery(TObject *Sender, bool &CanClose)
{
	int status = Ms3Recorder::Instance().SetMode(Ms3Recorder::MODE_IDLE);
	ReportError(status, "Failed to close MS3 recorder/player.");
	
	delete display;
	if (device)
	{
		device->SaveSettings(&device->settings);
		device->Disconnect();
		delete device;
		device = NULL;
	}

	filterChain.Release();

	AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".ini" );
	appSettings.bWindowMaximized = (this->WindowState == wsMaximized);
	if (!appSettings.bWindowMaximized)
	{
		// these values are meaningless is wnd is maximized
		appSettings.iWidth = this->Width;
		appSettings.iHeight = this->Height;
		appSettings.iPosY = this->Top;
		appSettings.iPosX = this->Left;
	}

	appSettings.Statistics.iPosX = frmStatistics->Left;
	appSettings.Statistics.iPosY = frmStatistics->Top;
	appSettings.Statistics.bVisible = frmStatistics->Visible;

	appSettings.Write(asConfigFile);

	CanClose = true;
}
//---------------------------------------------------------------------------

void TfrmMain::LoadDevice(AnsiString dllname)
{
    // get rid of artifacts first
	frmMain->Repaint();
	display->Update(Display::UPDATE_ALL);
	frmWait->Show("Loading input device dll, please wait...");

//	ToolBarMode->Enabled = false;
	if (device)
	{
		device->SaveSettings(&device->settings);
		device->Disconnect();
		delete device;
		device = NULL;
	}
	dataOsc.clear();
	dataFFT.clear();
	iStoredTriggerTypePos = -1;
	device = new DeviceInterface(dllname);
	if (device->Load())
	{
		ShowMessage("Failed loading device dll.");
		delete device;
		device = NULL;
		frmWait->Close();
		return;
	}
	device->callbackData = CallbackDataFn;
	device->callbackConnect = CallbackConnectFn;
	device->callbackTrigger = CallbackTriggerFn;

	display->SetDiv(device->capabilities->fSamplingPeriod[0]);
	SetCapabilities(device->capabilities);

	if (device->GetSettings(&device->settings) == 0)
	{
		trbarSensitivity->Position = device->settings.iSensitivity;
		trbarSampling->Position = device->settings.iSamplingPeriod;
		trbarTriggerLevel->Position = device->settings.iTriggerLevel;
		for (unsigned int i=0; i<device->capabilities->iTriggerTypeCount; i++)
		{
			if (device->capabilities->iTriggerTypes[i] == device->settings.eTriggerType)
			{
				cbTriggerType->ItemIndex = i;
				break;
			}
		}
		for (unsigned int i=0; i<device->capabilities->iCouplingTypeCount; i++)
		{
			if (device->capabilities->iCouplingType[i] == device->settings.eCouplingType)
			{
				cbCouplingType->ItemIndex = i;
				break;
			}
		}
		cbBufferSize->ItemIndex = cbBufferSize->Items->IndexOf(device->settings.iBufferSize);
	}

	ActionCalibrateSensitivity->Enabled = device->hasCalibration();
	calibration.clear();
	if (device->hasCalibration())
	{
		for (unsigned int i=0; i<device->capabilities->iSensitivityLevelsCount; i++)
		{
			struct DEV_CALIBRATION cal = {0, 1.0};
			device->GetCalibration(i, &cal.offset, &cal.gain);
			calibration.push_back(cal);
		}
		//ApplyCalibration();
	} else {
		for (unsigned int i=0; i<device->capabilities->iSensitivityLevelsCount; i++)
		{
			struct DEV_CALIBRATION cal = {0, 1.0};
			calibration.push_back(cal);
		}
	}

	frmWait->Show("Connecting to input device...");

	if (device->Connect())
	{
		frmWait->Close();
		// update only GUI state if required
		UpdateTriggerType(false);
		UpdateSensitivity(false);
		UpdateSampling(false);
		ShowMessage("Failed connecting to device.");
	} else
	{
		UpdateTriggerType();
		UpdateCouplingType();
		UpdateBufferSize();
		UpdateSensitivity();
		UpdateSampling();
		UpdateTriggerLevel();
//		ToolBarMode->Enabled = true;
		frmWait->Close();
	}
	current = upcoming;
}

std::auto_ptr<Graphics::TBitmap> TfrmMain::CreatePlotBitmap(void)
{
	bool bStatState = frmStatistics->Visible;
	if (bStatState)
	{
        // hide statistics window
		frmStatistics->Visible = false;
		display->Update(Display::UPDATE_ALL);
	}
	frmStatistics->Update();
	std::auto_ptr<Graphics::TBitmap> bitmap = display->CreateBitmap();
	if (bStatState)
	{
		frmStatistics->Visible = bStatState;
		this->SetFocus();
	}
	return bitmap;
}

void __fastcall TfrmMain::ActionCopyPlotClipbrdExecute(TObject *Sender)
{
	// make sure display is repainted
	display->Update(Display::UPDATE_ALL);
	std::auto_ptr<Graphics::TBitmap> bitmap = CreatePlotBitmap();
	TClipboard *pCB = Clipboard();
	pCB->Assign(&(*bitmap));
	(void)pCB;	// avoid unused variable warning
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateZoomDisplay(void)
{
	lblZoomX->Caption = (AnsiString)"X: " + (ValueStringFormatter::ValToStr(DisplayState().fZoomX, 2) + "x").c_str();
	lblZoomY->Caption = (AnsiString)"Y: " + (ValueStringFormatter::ValToStr(DisplayState().fZoomY, 2) + "x").c_str();
}


void __fastcall TfrmMain::TrackbarKeyUp(TObject *Sender, WORD &Key,
	  TShiftState Shift)
{
	UpdateTrackBar(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::TrackBarMouseUp(TObject *Sender,
	TMouseButton Button, TShiftState Shift, int X, int Y)
{
	UpdateTrackBar(Sender);
}

void TfrmMain::UpdateTrackBar(TObject *Sender)
{
	if (Sender == trbarSensitivity)
	{
		UpdateSensitivity();
	}
	else if (Sender == trbarSampling)
	{
		UpdateSampling();
	}
	else if (Sender == trbarTriggerLevel)
	{
		UpdateTriggerLevel();
	}
}

void __fastcall TfrmMain::miStatisticsClick(TObject *Sender)
{
	frmStatistics->Show();
	ScopedLock<Mutex> lock(mutexDeviceData);

	unsigned int iExpectedFrameSize = device->capabilities->iBufferSizes[cbBufferSize->ItemIndex];
	if (dataOsc.size() == iExpectedFrameSize)
	{
		if (display->GetMode() == Display::MODE_RECORDER)
			frmStatistics->Process(dataRec, current.fDivX);
		else
			frmStatistics->Process(dataOsc, current.fDivX);
	}
}
//---------------------------------------------------------------------------

void TfrmMain::FirstRunDialog(void)
{
	const char msg[] = "No input device dll is selected.\n"
		"It seems like application is running for ther first time\n"
		"or configuration has been removed.\n"
		"Selecting device library is REQUIRED.";
	MessageBox(this->Handle, msg, "No input device plugin selected", MB_OK | MB_ICONINFORMATION);
	    ActionShowCommonSettingsExecute(NULL);
	}

void __fastcall TfrmMain::ActionShowCommonSettingsExecute(TObject *Sender)
{
	AnsiString asPrevDevice = appSettings.asInputDeviceDllName;
	frmSettings->appSettings = &appSettings;
	frmSettings->ShowModal();
	if (appSettings.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
	if (appSettings.asInputDeviceDllName != asPrevDevice)
		LoadDevice(asDevDllDir + "\\" + appSettings.asInputDeviceDllName);
	Timer->Interval = 1000/appSettings.iDisplayRefreshFreq;
	
	display->SetVisibility(Display::MODE_OSCILLOSCOPE, appSettings.GraphOscilloscope.bGridX,
		appSettings.GraphOscilloscope.bGridY, appSettings.GraphOscilloscope.bGridLabels);
	display->SetVisibility(Display::MODE_SPECTRUM, appSettings.GraphSpectrum.bGridX,
		appSettings.GraphSpectrum.bGridY, appSettings.GraphSpectrum.bGridLabels);
	display->SetVisibility(Display::MODE_RECORDER, appSettings.GraphRecorder.bGridX,
		appSettings.GraphRecorder.bGridY, appSettings.GraphRecorder.bGridLabels);
	frmPlotToolbar->SetTransparecy(appSettings.GraphToolboxTransparency);
	FFT::SetWindowType(appSettings.FFT.asWindowName);
	bPaintUpdate = true;
}
//---------------------------------------------------------------------------

void TfrmMain::SetMode(enum Display::E_DISPLAY_MODE mode)
{
	int oldmode = -1;

    // restore trigger type to state before switching to recorder mode
	if (mode != oldmode && oldmode == Display::MODE_RECORDER)
	{
		if (cbTriggerType->Items->Count > iStoredTriggerTypePos && iStoredTriggerTypePos != -1)
		{
			cbTriggerType->ItemIndex = iStoredTriggerTypePos;		
			UpdateTriggerType();
		}
	}

	bDataUpdate = true;
	if (mode == Display::MODE_OSCILLOSCOPE)
		display->data = &dataOsc;
	else if (mode == Display::MODE_SPECTRUM)
		display->data = &dataFFT;
	else if (mode == Display::MODE_RECORDER)
	{
		display->data = &dataRec;
		iStoredTriggerTypePos = cbTriggerType->ItemIndex;
		for (unsigned int i=0; i<device->capabilities->iTriggerTypeCount; i++)
		{
			if (device->capabilities->iTriggerTypes[i] == TRIGGER_MANUAL)
			{
				cbTriggerType->ItemIndex = i;
				break;
			}
		}
		UpdateTriggerType();
	}
	display->SetMode(mode);
	UpdateZoomDisplay();


	if (mode != Display::MODE_RECORDER)
	{
		cbTriggerType->Enabled = true;
		if (device->capabilities->iTriggerTypes[cbTriggerType->ItemIndex] == TRIGGER_MANUAL)
			btnManualTrigger->Visible = true;
		btnRecorderStart->Visible = false;
		btnRecorderStop->Visible = false;
		btnRecorderClear->Visible = false;
		lblInterval->Visible = false;
		edInterval->Visible = false;
		lblSeconds->Visible = false;

		tmrRecorder->Enabled = false;
		btnRecorderStart->Enabled = true;
		btnRecorderStop->Enabled = false;
	}
	else
	{
		cbTriggerType->Enabled = false;
		btnManualTrigger->Visible = false;
		btnRecorderStart->Visible = true;
		btnRecorderStop->Visible = true;
		btnRecorderClear->Visible = true;
		lblInterval->Visible = true;
		edInterval->Visible = true;
		lblSeconds->Visible = true;
	}	
}



void __fastcall TfrmMain::btnRecorderStartClick(TObject *Sender)
{
	float interval = StrToFloatDef(edInterval->Text, -1.0);
	if (interval < Settings::fRecorderMinInterval || interval > Settings::fRecorderMaxInterval)
	{
		MessageBox(this->Handle, "Recorder interval outside allowed range!", this->Caption.c_str(),
			MB_ICONEXCLAMATION);
		//edInterval->Text = AnsiString::FloatToStrF(appSettings.fRecorderInterval, AnsiString::sffGeneral, 1, 3);
		return;
	}
	appSettings.fRecorderInterval = interval;

	tmrRecorder->Interval = interval * 1000.0;


	tmrRecorder->Enabled = true;
	btnRecorderStart->Enabled = false;
	edInterval->Enabled = false;
	btnRecorderStop->Enabled = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnRecorderStopClick(TObject *Sender)
{
	tmrRecorder->Enabled = false;
	btnRecorderStart->Enabled = true;
	edInterval->Enabled = true;
	btnRecorderStop->Enabled = false;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnRecorderClearClick(TObject *Sender)
{
	ScopedLock<Mutex> lock(mutexDeviceData);
	dataRec.clear();
	bDataUpdate = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::tmrRecorderTimer(TObject *Sender)
{
	assert(display->GetMode() == Display::MODE_RECORDER);
	// execute manual trigger
	if (device)
		device->ManualTrigger();
	/** \todo check if previous sample already received */
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionExportPlotBitmapExecute(TObject *Sender)
{
	SaveDialog->DefaultExt = "bmp";
	SaveDialog->Filter = "Bitmap (*.bmp)|*.bmp";
	SaveDialog->FilterIndex = 1;
	if (DirectoryExists(appSettings.asPlotPictureExportDir))
		SaveDialog->InitialDir = appSettings.asPlotPictureExportDir;
	SaveDialog->FileName = SaveDialog->InitialDir +
		((SaveDialog->InitialDir!="")?("\\"):("")) + FormatDateTime("yyyymmdd_hhnnss", Now());
	if (SaveDialog->Execute())
	{
		// refresh display (save dialog may be still visible on low spec PC)
		display->Update(Display::UPDATE_ALL);
		std::auto_ptr<Graphics::TBitmap> bitmap = CreatePlotBitmap();
		bitmap->PixelFormat = pf8bit;
		bitmap->SaveToFile(SaveDialog->FileName);
		appSettings.asPlotPictureExportDir = ExtractFileDir(SaveDialog->FileName);
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionCalibrateSensitivityExecute(TObject *Sender)
{
	if (!frmCalibrateSensitivity->Visible)
		frmCalibrateSensitivity->Show();
	frmCalibrateSensitivity->BringToFront();
}
//---------------------------------------------------------------------------

int TfrmMain::CallbackCalculateCalibrationFn(float fVoltageHigh, int* const offset, float* const gain)
{
	float fZero, fCal, dummy;
	int status = display->GetCursorPhysical(Display::CURSOR2, dummy, fZero);
	status |= display->GetCursorPhysical(Display::CURSOR1, dummy, fCal);
	if (status)
		return status;
	// calculate offset
	int bitsZero = fZero/(device->capabilities->fSensitivityLevels[trbarSensitivity->Position]*current.calibration.gain) +
		current.calibration.offset;
	*offset = 0.5 + bitsZero;

	// calculate gain
	int bitsCal = fCal/(device->capabilities->fSensitivityLevels[trbarSensitivity->Position]*current.calibration.gain);
	bitsCal -= *offset;
	*gain = fVoltageHigh /
		( bitsCal * device->capabilities->fSensitivityLevels[trbarSensitivity->Position] );
	return 0;
}

int TfrmMain::CallbackStoreCalibrationFn(const int offset, const float gain)
{
	calibration[trbarSensitivity->Position].offset = offset;
	calibration[trbarSensitivity->Position].gain = gain;
	device->StoreCalibration(trbarSensitivity->Position , offset, gain);
	ApplyCalibration();
	return 0;
}

void TfrmMain::CallbackPlotToolbarClickFn(int tag)
{
	// map tag to button
	TToolButton *Button;
	switch (tag)
	{
	case 1:
		Button = btnMouseCursor;
		break;
	case 2:
		Button = btnMouseMove;
		break;
	case 3:
		Button = btnMouseZoomXY;
		break;
	case 4:
		Button = btnMouseZoomX;
		break;
	case 5:
		Button = btnMouseZoomY;
		break;
	case 6:
		Button = btnZoomClear;
		break;
	default:
		return;
	}
	Button->Action->Execute();
	if (Button->Grouped)
		Button->Down = true;
}

void TfrmMain::ApplyCalibration(void)
{
	UpdateSensitivity();
}

void __fastcall TfrmMain::ActionExportDataExecute(TObject *Sender)
{
	SaveDialog->DefaultExt = "json";
	SaveDialog->Filter = "JSON files (*.json)|*.json|CSV files (*.csv)|*.csv";
	SaveDialog->FilterIndex = 1;
	if (DirectoryExists(appSettings.asPlotDataExportDir))
		SaveDialog->InitialDir = appSettings.asPlotDataExportDir;
	SaveDialog->FileName = SaveDialog->InitialDir +
		((SaveDialog->InitialDir!="")?("\\"):("")) + FormatDateTime("yyyymmdd_hhnnss", Now());

	enum Display::E_DISPLAY_MODE mode = display->GetMode();
	std::vector<float> *vdata;
	if (mode == Display::MODE_RECORDER)
		vdata = &dataRec;
	else if (mode == Display::MODE_OSCILLOSCOPE || mode == Display::MODE_SPECTRUM)
		vdata = &dataOsc;
	else
		return;

	if (SaveDialog->Execute())
	{
		enum Recorder::E_FMT format;
		if (SaveDialog->FilterIndex == 1)
			format = Recorder::FMT_JSON;
		else
			format = Recorder::FMT_CSV;
		int status = Recorder::Store(SaveDialog->FileName.c_str(), format,
			*vdata, current.timestamp, current.fDivX, current.coupling, current.trigger_type);
		appSettings.asPlotDataExportDir = ExtractFileDir(SaveDialog->FileName);

		if (status)
			MessageBox(this->Handle, "Failed saving file", this->Caption.c_str(), MB_ICONEXCLAMATION);
	}
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::FormShow(TObject *Sender)
{
	static bool once = false;
	if (!once)
	{
		once = true;
		// frmLog is already created - set callback to preview log entries
		CLog::Instance()->callbackLog = frmLog->OnLog;
		frmLog->callbackClose = CallbackCloseLogFn;

		frmCalibrateSensitivity->callbackCalculate = CallbackCalculateCalibrationFn;
		frmCalibrateSensitivity->callbackStore = CallbackStoreCalibrationFn;

		display->SetVisibility(Display::MODE_OSCILLOSCOPE, appSettings.GraphOscilloscope.bGridX,
			appSettings.GraphOscilloscope.bGridY, appSettings.GraphOscilloscope.bGridLabels);
		display->SetVisibility(Display::MODE_SPECTRUM, appSettings.GraphSpectrum.bGridX,
			appSettings.GraphSpectrum.bGridY, appSettings.GraphSpectrum.bGridLabels);
		display->SetVisibility(Display::MODE_RECORDER, appSettings.GraphRecorder.bGridX,
			appSettings.GraphRecorder.bGridY, appSettings.GraphRecorder.bGridLabels);

		frmPlotToolbar->SetTransparecy(appSettings.GraphToolboxTransparency);

		// load device dlls
		asDevDllDir = ExtractFileDir(Application->ExeName) + "\\device";
		DeviceInterface::EnumerateDlls(asDevDllDir);

		Timer->Interval = 1000/appSettings.iDisplayRefreshFreq;
		display->ChangeSize(pnlCenter->Width, pnlCenter->Height);

		if (appSettings.asInputDeviceDllName != "")
		{
			AnsiString dllname = asDevDllDir + "\\" + appSettings.asInputDeviceDllName;
			for (unsigned int i=0; i<DeviceInterface::dlls.size(); i++)
			{
				if (DeviceInterface::dlls[i].name == dllname)
				{
					LoadDevice(dllname);
					break;
				}
			}
		}

		if (!device)
		{
			FirstRunDialog();
		}
		if (!device)
		{
			const char msg[] = "No input device dll is selected.\n"
				"Selecting input plugin is REQUIRED.\n"
				"Note: device plugins should be placed in \\device subdirectory of application.";
			MessageBox(this->Handle, msg, "Execution aborted", MB_ICONSTOP);
			Close();
			return;
		}
		SetMode(Display::MODE_OSCILLOSCOPE);
		frmPlotToolbar->callbackClick = CallbackPlotToolbarClickFn;

		frmLog->SetLogLinesLimit(appSettings.iMaxUiLogLines);

		frmStatistics->Left = appSettings.Statistics.iPosX;
		frmStatistics->Top = appSettings.Statistics.iPosY;
		if (appSettings.Statistics.bVisible)
			frmStatistics->Show();		
	}	
}
//---------------------------------------------------------------------------

/** \brief Set diaplay Y axis zoom & offset to fit trace into
*/
void __fastcall TfrmMain::ActionAutoFitYExecute(TObject *Sender)
{
	std::vector<float> *data;
	if (display->GetMode() == Display::MODE_OSCILLOSCOPE)
		data = &dataOsc;
	else if (display->GetMode() == Display::MODE_SPECTRUM)
		data = &dataFFT;
	else if (display->GetMode() == Display::MODE_RECORDER)
		data = &dataRec;
	else
	{
		ShowMessage("unimplemented");
		return;
	}
	if (!data || !data->size())
		return;
	float fMin = std::numeric_limits<float>::max();
	float fMax = (-1.0) * fMin;

	std::vector<float>::const_iterator it;
	for (it = data->begin(); it != data->end(); ++it)
	{
		float val = *it;
		if (val < fMin)
			fMin = val;
		if (val > fMax)
			fMax = val;
	}

	if (fMin == fMax)
		return;

	float fDivVal = (float)pow(2, device->capabilities->iBitsPerSample) / (float)10 *
		device->capabilities->fSensitivityLevels[trbarSensitivity->Position] *
		calibration[trbarSensitivity->Position].gain;

	DisplayState().fZoomY = fDivVal*10.0/(fMax-fMin);
	display->SetZoom(Display::OFFSET_LEAVE, Display::OFFSET_LEAVE, DisplayState().fZoomX, DisplayState().fZoomY);
	display->SetOffsetY((fMax + fMin)/2.0);
	UpdateZoomDisplay();
	bPaintUpdate = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionAutoFitXExecute(TObject *Sender)
{
	std::vector<float> *data;
	if (display->GetMode() == Display::MODE_OSCILLOSCOPE)
		data = &dataOsc;
	else if (display->GetMode() == Display::MODE_SPECTRUM)
		data = &dataFFT;
	else if (display->GetMode() == Display::MODE_RECORDER)
		data = &dataRec;
	else
	{
		ShowMessage("unimplemented");
		return;
	}
	if (!data || !data->size())
		return;
	DisplayState().fZoomX = (float)pnlCenter->Width / data->size();

	display->SetZoom(Display::OFFSET_CLEAR, Display::OFFSET_LEAVE, DisplayState().fZoomX, DisplayState().fZoomY);
	UpdateZoomDisplay();
	bPaintUpdate = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionLoadJsonFileExecute(TObject *Sender)
{
	OpenDialog->DefaultExt = "json";
	OpenDialog->Filter = "JSON files (*.json)|*.json|All files (*.*)|*.*";
	OpenDialog->FilterIndex = 1;
	if (DirectoryExists(appSettings.asPlotDataExportDir))
		OpenDialog->InitialDir = appSettings.asPlotDataExportDir;
	if (OpenDialog->Execute())
	{
		int status = frmJsonViewer->LoadFile(OpenDialog->FileName);
		if (status != 0)
		{
			MessageBox(this->Handle, "Error loading JSON file",
				this->Caption.c_str(), MB_ICONEXCLAMATION);
			return;
		}
		bJsonViewerMode = true;
		ActionExportData->Enabled = false;
		dataOsc = frmJsonViewer->data.samples;
		display->SetDiv(frmJsonViewer->data.fDivX);
		display->Update(Display::UPDATE_ALL);
		pnlOscControl->Visible = false;
		frmJsonViewer->Parent = pnlRight;
		frmJsonViewer->Visible = true;
		frmJsonViewer->callbackClose = CallbackJsonViewerClose;
	}
}
//---------------------------------------------------------------------------

void TfrmMain::CallbackJsonViewerClose(void)
{
	bJsonViewerMode = false;
	ActionExportData->Enabled = true;
	pnlOscControl->Visible = true;
}

void TfrmMain::CallbackMs3ViewerClose(void)
{
	bMs3ViewerMode = false;
	ActionMs3Record->Enabled = true;
	ActionExportData->Enabled = true;
	pnlOscControl->Visible = true;
}

void TfrmMain::CallbackMs3ViewerDataUpdate(void)
{
	filterChain.Process(dataOsc);
	if (frmStatistics->Visible)
		frmStatistics->Process(dataOsc, fMs3DivX);
	if (display->GetMode() == Display::MODE_SPECTRUM)
	{
		CalculateSpectrum();
	}
	display->Update(Display::UPDATE_ALL);
}

void __fastcall TfrmMain::ActionShowDataFilterChainExecute(TObject *Sender)
{
	frmFilterSelector->ShowModal();
	if (frmFilterSelector->ModalResult == mrOk)
	{
		filterChain.Configure(frmFilterSelector->Get());
		bDataUpdate = true;
	}
}
//---------------------------------------------------------------------------

void TfrmMain::ReportError(int status, const char* const description)
{
	if (status == 0)
		return;
	AnsiString asMsg = (AnsiString)description + "\nError code: " + status;
	LOG(E_LOG_TRACE, "%s", asMsg.c_str());
	MessageBox(this->Handle, asMsg.c_str(), this->Caption.c_str(), MB_ICONEXCLAMATION);
}

void __fastcall TfrmMain::ActionMs3RecordExecute(TObject *Sender)
{
	if (ActionMs3Record->Checked)
	{
		SaveDialog->DefaultExt = "ms3";
		SaveDialog->Filter = "MS3 files (*.ms3)|*.ms3";
		SaveDialog->FilterIndex = 1;
		if (DirectoryExists(appSettings.asPlotDataExportDir))
			SaveDialog->InitialDir = appSettings.asPlotDataExportDir;
		SaveDialog->FileName = SaveDialog->InitialDir +
			((SaveDialog->InitialDir!="")?("\\"):("")) + FormatDateTime("yyyymmdd_hhnnss", Now());
		if (SaveDialog->Execute())
		{
			if (FileExists(SaveDialog->FileName))
			{
				if (MessageBox(this->Handle, "File already exists, file will be overwritten.\r\nAre you sure you want to continue?",
					this->Caption.c_str(), MB_YESNO | MB_DEFBUTTON2 | MB_ICONEXCLAMATION) != IDYES)
				{
                    ActionMs3Record->Checked = false;
					return;
				}
			}
			appSettings.asPlotDataExportDir = ExtractFileDir(SaveDialog->FileName);
			int status = Ms3Recorder::Instance().SetMode(Ms3Recorder::MODE_WRITE, SaveDialog->FileName);
			ReportError(status, "Failed to open MS3 file for writing");
			if (status)
				ActionMs3Record->Checked = false;
			else
				LOG(E_LOG_TRACE, "Recording data to %s file started", SaveDialog->FileName.c_str());
		}
		else
		{
			ActionMs3Record->Checked = false;
		}
	}
	else
	{
		int status = Ms3Recorder::Instance().SetMode(Ms3Recorder::MODE_IDLE);
		ReportError(status, "Failed to close MS3 recorder/player.");
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionMs3LoadExecute(TObject *Sender)
{
	if (ActionMs3Record->Checked)
		ActionMs3Record->Execute();
		
	OpenDialog->DefaultExt = "ms3";
	OpenDialog->Filter = "MS3 files (*.ms3)|*.ms3|All files (*.*)|*.*";
	OpenDialog->FilterIndex = 1;
	if (DirectoryExists(appSettings.asPlotDataExportDir))
		OpenDialog->InitialDir = appSettings.asPlotDataExportDir;
	if (OpenDialog->Execute())
	{
		int status = frmMs3Viewer->LoadFile(OpenDialog->FileName, CallbackMs3ViewerDataUpdate, &fMs3DivX, &dataOsc);
		if (status != 0)
		{
			MessageBox(this->Handle, "Error loading MS3 file",
				this->Caption.c_str(), MB_ICONEXCLAMATION);
			bMs3ViewerMode = false;
			ActionMs3Record->Enabled = true;
			return;
		}
		bMs3ViewerMode = true;
		ActionMs3Record->Enabled = false;

		pnlOscControl->Visible = false;
		frmMs3Viewer->Parent = pnlRight;
		frmMs3Viewer->Visible = true;
		frmMs3Viewer->callbackClose = CallbackMs3ViewerClose;
	}
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::ActionMs3FileRepairExecute(TObject *Sender)
{
	OpenDialog->DefaultExt = "ms3";
	OpenDialog->Filter = "MS3 files (*.ms3)|*.ms3|All files (*.*)|*.*";
	OpenDialog->FilterIndex = 1;
	if (DirectoryExists(appSettings.asPlotDataExportDir))
		OpenDialog->InitialDir = appSettings.asPlotDataExportDir;
	if (OpenDialog->Execute())
	{
		enum TfrmMs3FileRepair::E_STATUS status;
		frmMs3FileRepair->Execute(OpenDialog->FileName, status);
		MessageBox(this->Handle, frmMs3FileRepair->GetStatusDescription(status).c_str(),
			this->Caption.c_str(), MB_ICONINFORMATION);
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::edIntervalExit(TObject *Sender)
{
	float interval = StrToFloatDef(edInterval->Text, -1.0);
	if (interval < Settings::fRecorderMinInterval || interval > Settings::fRecorderMaxInterval)
	{
		MessageBox(this->Handle, "Recorder interval outside allowed range!", this->Caption.c_str(),
			MB_ICONEXCLAMATION);
	}
	else
	{
		appSettings.fRecorderInterval = interval;
	}
}
//---------------------------------------------------------------------------

