/** \file
 *  \brief Device capabilities specification. This file is shared by miniscope and device dll projects.
 */

#ifndef MiniscopeDeviceCapabilitiesH
#define MiniscopeDeviceCapabilitiesH

/** \brief Supported triggering modes
 */
enum E_TRIGGER_TYPE
{
	TRIGGER_MANUAL = 0,    ///< "manual" trigger per request
    TRIGGER_SLOPE_UP,       ///< raising slope
	TRIGGER_SLOPE_DOWN,     ///< falling slope
	TRIGGER_CONTINUOUS		///< non-stop trigger (or no trigger at all)
};

enum E_ACQUISITION_MODE
{
	ACQUISITION_1SHOT = 0, ///< stop after acquiring first data buffer
	ACQUISITION_CONTINUOUS ///< allow continuous triggering
};

/** \brief Supported input coupling modes
 */
enum E_COUPLING_TYPE
{
    COUPLING_GND = 0,       ///< input tied to ground
    COUPLING_DC,            ///< direct current coupling
    COUPLING_AC             ///< DC path blocked
};

/** \brief Device capabilities structure
*/
struct S_SCOPE_CAPABILITIES
{
    unsigned int iBitsPerSample;        ///< number of bits per data sample
    unsigned int iSensitivityLevelsCount;   ///< number of available sensitivity level settings
	float *fSensitivityLevels;          ///< ptr to array of sensitivity per LSB values (array size defined by iSensitivityLevelsCount)
										///< values must be ascending
    unsigned int iSamplingPeriodCount;  ///< number of available sampling period time settings
    float *fSamplingPeriod;      		///< ptr to sampling period array [s]

	unsigned int iTriggerTypeCount;     ///< number of available trigger types
	enum E_TRIGGER_TYPE *iTriggerTypes; ///< ptr to trigger types array

	unsigned int iAcqModeCount;         ///< number of available acquisition modes
	enum E_ACQUISITION_MODE *iAcqModeTypes;    ///< ptr to acquisition modes array

	unsigned int iCouplingTypeCount;   ///< number of available coupling modes
	enum E_COUPLING_TYPE *iCouplingType;        ///< ptr to available coupling modes array

	unsigned int iBufferSizeCount;      ///< number of selectable acquisition buffer sizes
	unsigned int *iBufferSizes;         ///< ptr to acquisition buffer sizes array

    unsigned int iOffset;               ///< is offset setting available?
};

/** \brief Device settings structure
*/
struct S_SCOPE_SETTINGS
{
	unsigned int iSensitivity;
	unsigned int iSamplingPeriod;
	enum E_TRIGGER_TYPE eTriggerType;
	int iTriggerLevel;
	enum E_ACQUISITION_MODE eAcqModeType;
	enum E_COUPLING_TYPE eCouplingType;
	unsigned int iBufferSize;
	unsigned int iOffset;
};

#endif // MiniscopeDeviceCapabilitiesH
